"""
Classes for calling functions a schedule. Has time zone support.

For example, to run a job at 08:00 every morning in 'Asia/Calcutta':

>>> import zoneinfo
>>> job = lambda: print("time is now", datetime.datetime())
>>> time = datetime.time(8, tzinfo=zoneinfo.ZoneInfo('Asia/Calcutta'))
>>> cmd = PeriodicCommandFixedDelay.daily_at(time, job)
>>> print(cmd)
PeriodicCommandFixedDelay: <function <lambda> at ...> at 2...T08:00:00+05:30
>>> sched = InvokeScheduler()
>>> sched.add(cmd)
>>> while True:  # doctest: +SKIP
...     sched.run_pending()
...     time.sleep(.1)

By default, the scheduler uses timezone-aware times in UTC. A
client may override the default behavior by overriding ``now``
and ``from_timestamp`` functions.

>>> now()
datetime.datetime(...utc)
>>> from_timestamp(1718723533.7685602)
datetime.datetime(...utc)

Consumers may supply custom attributes to a command and those
should be retained.

>>> cmd = PeriodicCommandFixedDelay.daily_at(time, job)
>>> cmd.name = 'my task name'
>>> cmd.next().name
'my task name'

Alternatively, to associate custom context (such as a name) with
a command, implement the target as a callable class whose
``__str__`` describes it. The name then naturally carries through
``next()`` as part of the target itself:

>>> class NamedTask:
...     def __init__(self, name, func):
...         self.name = name
...         self.func = func
...     def __str__(self):
...         return self.name
...     def __call__(self):
...         return self.func()
>>> cmd = PeriodicCommandFixedDelay.daily_at(time, NamedTask('my task name', job))
>>> print(cmd)
PeriodicCommandFixedDelay: my task name at 2...T08:00:00+05:30

To bind arguments to the target function, use ``functools.partial``:

>>> import functools
>>> def greet(whom): print(f"Hello, {whom}!")
>>> cmd = PeriodicCommandFixedDelay.daily_at(time, functools.partial(greet, 'world'))
>>> cmd.target()
Hello, world!
"""

from __future__ import annotations

import abc
import bisect
import collections.abc
import datetime
import numbers
from typing import TYPE_CHECKING, Any

from jaraco.collections import set_defaults
from jaraco.context import suppress
from jaraco.functools import passthrough  # type: ignore[attr-defined]

from .utc import fromtimestamp as from_timestamp
from .utc import now as now

if TYPE_CHECKING:
    from typing_extensions import Self


class DelayedCommand(datetime.datetime):
    """
    A command to be executed after some delay (seconds or timedelta).
    """

    delay: datetime.timedelta = datetime.timedelta()
    target: Any  # Expected type depends on the scheduler used

    @classmethod
    def from_datetime(cls, other: datetime.datetime) -> Self:
        return cls(
            other.year,
            other.month,
            other.day,
            other.hour,
            other.minute,
            other.second,
            other.microsecond,
            other.tzinfo,
        )

    @classmethod
    def after(cls, delay: datetime.timedelta | float, target: Any) -> Self:
        if not isinstance(delay, datetime.timedelta):
            delay = datetime.timedelta(seconds=delay)
        due_time = now() + delay
        cmd = cls.from_datetime(due_time)
        cmd.delay = delay
        cmd.target = target
        return cmd

    @staticmethod
    def _from_timestamp(input: datetime.datetime | float) -> datetime.datetime:
        """
        If input is a real number, interpret it as a Unix timestamp
        (seconds sinc Epoch in UTC) and return a timezone-aware
        datetime object. Otherwise return input unchanged.
        """
        if isinstance(input, datetime.datetime):
            return input
        return from_timestamp(input)

    @classmethod
    def at_time(cls, at: datetime.datetime | float, target: Any) -> Self:
        """
        Construct a DelayedCommand to come due at `at`, where `at` may be
        a datetime or timestamp.
        """
        at = cls._from_timestamp(at)
        cmd = cls.from_datetime(at)
        cmd.delay = at - now()
        cmd.target = target
        return cmd

    def due(self) -> bool:
        return now() >= self

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: {self.target} at {self.isoformat()}"


class PeriodicCommand(DelayedCommand):
    """
    Like a delayed command, but expect this command to run every delay
    seconds.
    """

    def _next_time(self) -> Self:
        """
        Add delay to self, localized
        """
        return self + self.delay

    @passthrough  # type: ignore[untyped-decorator]
    @suppress(TypeError)
    def _reflect(self, other: Any) -> Self:  # type: ignore[return]
        """
        Ensure any custom attributes from other are present on self.
        """
        set_defaults(vars(self), **vars(other))

    def next(self) -> Self:
        cmd = self.__class__.from_datetime(self._next_time())
        cmd.delay = self.delay
        cmd.target = self.target
        reflected: Self = cmd._reflect(self)
        return reflected

    def __setattr__(self, key: str, value: Any) -> None:
        if key == 'delay' and not value > datetime.timedelta():
            raise ValueError("A PeriodicCommand must have a positive, non-zero delay.")
        super().__setattr__(key, value)


class PeriodicCommandFixedDelay(PeriodicCommand):
    """
    Like a periodic command, but don't calculate the delay based on
    the current time. Instead use a fixed delay following the initial
    run.
    """

    @classmethod
    def at_time(  # type: ignore[override] # jaraco/tempora#39
        cls,
        at: datetime.datetime | float,
        delay: datetime.timedelta | numbers.Number,
        target: Any,
    ) -> Self:
        """
        >>> cmd = PeriodicCommandFixedDelay.at_time(0, 30, None)
        >>> cmd.delay.total_seconds()
        30.0
        """
        at = cls._from_timestamp(at)
        cmd = cls.from_datetime(at)
        if isinstance(delay, numbers.Number):
            delay = datetime.timedelta(seconds=delay)  # type: ignore[arg-type] # python/mypy#3186#issuecomment-1571512649
        cmd.delay = delay
        cmd.target = target
        return cmd

    @classmethod
    def daily_at(cls, at: datetime.time, target: Any) -> Self:
        """
        Schedule a command to run at a specific time each day.

        >>> from tempora import utc
        >>> noon = utc.time(12, 0)
        >>> cmd = PeriodicCommandFixedDelay.daily_at(noon, None)
        >>> cmd.delay.total_seconds()
        86400.0
        """
        daily = datetime.timedelta(days=1)
        # convert when to the next datetime matching this time
        when = datetime.datetime.combine(datetime.date.today(), at)
        when -= daily
        while when < now():
            when += daily
        return cls.at_time(when, daily, target)


class Scheduler:
    """
    A rudimentary abstract scheduler accepting DelayedCommands
    and dispatching them on schedule.
    """

    def __init__(self) -> None:
        self.queue: list[DelayedCommand] = []

    def add(self, command: DelayedCommand) -> None:
        bisect.insort(self.queue, command)

    def run_pending(self) -> None:
        while self.queue:
            command = self.queue[0]
            if not command.due():
                break
            self.run(command)
            if isinstance(command, PeriodicCommand):
                self.add(command.next())
            del self.queue[0]

    @abc.abstractmethod
    def run(self, command: DelayedCommand) -> None:
        """
        Run the command
        """


class InvokeScheduler(Scheduler):
    """
    Command targets are functions to be invoked on schedule.
    """

    def run(self, command: DelayedCommand) -> None:
        command.target()


class CallbackScheduler(Scheduler):
    """
    Command targets are passed to a dispatch callable on schedule.
    """

    def __init__(self, dispatch: collections.abc.Callable[..., Any]) -> None:
        super().__init__()
        self.dispatch = dispatch

    def run(self, command: DelayedCommand) -> None:
        self.dispatch(command.target)



from __future__ import annotations

import collections.abc
import contextlib
import datetime
import functools
import numbers
import time
from types import TracebackType
from typing import TYPE_CHECKING, Any

import jaraco.functools

if TYPE_CHECKING:
    from typing_extensions import Self


class Stopwatch:
    """
    A simple stopwatch that starts automatically.

    >>> w = Stopwatch()
    >>> _1_sec = datetime.timedelta(seconds=1)
    >>> w.split() < _1_sec
    True
    >>> time.sleep(1.0)
    >>> w.split() >= _1_sec
    True
    >>> w.stop() >= _1_sec
    True
    >>> w.reset()
    >>> w.start()
    >>> w.split() < _1_sec
    True

    Launch the Stopwatch in a context:

    >>> with Stopwatch() as watch:
    ...     assert isinstance(watch.split(), datetime.timedelta)

    After exiting the context, the watch is stopped; read the
    elapsed time directly:

    >>> watch.elapsed
    datetime.timedelta(...)
    >>> watch.elapsed.seconds
    0
    """

    def __init__(self) -> None:
        self.reset()
        self.start()

    def reset(self) -> None:
        self.elapsed = datetime.timedelta(0)
        with contextlib.suppress(AttributeError):
            del self._start

    def _diff(self) -> datetime.timedelta:
        return datetime.timedelta(seconds=time.monotonic() - self._start)

    def start(self) -> None:
        self._start = time.monotonic()

    def stop(self) -> datetime.timedelta:
        self.elapsed += self._diff()
        del self._start
        return self.elapsed

    def split(self) -> datetime.timedelta:
        return self.elapsed + self._diff()

    # context manager support
    def __enter__(self) -> Self:
        self.start()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self.stop()


class IntervalGovernor:
    """
    Decorate a function to only allow it to be called once per
    min_interval. Otherwise, it returns None.

    >>> gov = IntervalGovernor(30)
    >>> gov.min_interval.total_seconds()
    30.0
    """

    def __init__(self, min_interval: datetime.timedelta | numbers.Number) -> None:
        if isinstance(min_interval, numbers.Number):
            min_interval = datetime.timedelta(seconds=min_interval)  # type: ignore[arg-type] # python/mypy#3186#issuecomment-1571512649
        self.min_interval = min_interval
        self.last_call: Stopwatch | None = None

    def decorate(
        self, func: collections.abc.Callable[..., Any]
    ) -> collections.abc.Callable[..., Any]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            allow = not self.last_call or self.last_call.split() > self.min_interval
            if allow:
                self.last_call = Stopwatch()
                return func(*args, **kwargs)

        return wrapper

    __call__ = decorate


class Timer(Stopwatch):
    """
    Watch for a target elapsed time.

    >>> t = Timer(0.1)
    >>> t.expired()
    False
    >>> __import__('time').sleep(0.15)
    >>> t.expired()
    True
    """

    def __init__(
        self, target: float | datetime.timedelta | None = float('Inf')
    ) -> None:
        self.target = self._accept(target)
        super().__init__()

    @staticmethod
    def _accept(target: float | datetime.timedelta | None) -> float:
        """
        Accept None or ∞ or datetime or numeric for target

        >>> Timer._accept(datetime.timedelta(seconds=30))
        30.0
        >>> Timer._accept(None)
        inf
        """
        if isinstance(target, datetime.timedelta):
            target = target.total_seconds()

        if target is None:
            # treat None as infinite target
            target = float('Inf')

        return target

    def expired(self) -> bool:
        return self.split().total_seconds() > self.target


class BackoffDelay(collections.abc.Iterator[int | float]):
    """
    Exponential backoff delay.

    Useful for defining delays between retries. Consider for use
    with ``jaraco.functools.retry_call`` as the cleanup.

    Default behavior has no effect; a delay or jitter must
    be supplied for the call to be non-degenerate.

    >>> bd = BackoffDelay()
    >>> bd()
    >>> bd()

    The following instance will delay 10ms for the first call,
    20ms for the second, etc.

    >>> bd = BackoffDelay(delay=0.01, factor=2)
    >>> bd()
    >>> bd()

    Inspect and adjust the state of the delay anytime.

    >>> bd.delay
    0.04
    >>> bd.delay = 0.01

    Set limit to prevent the delay from exceeding bounds.

    >>> bd = BackoffDelay(delay=0.01, factor=2, limit=0.015)
    >>> bd()
    >>> bd.delay
    0.015

    To reset the backoff, simply call ``.reset()``:

    >>> bd.reset()
    >>> bd.delay
    0.01

    Iterate on the object to retrieve/advance the delay values.

    >>> next(bd)
    0.01
    >>> next(bd)
    0.015
    >>> import itertools
    >>> tuple(itertools.islice(bd, 3))
    (0.015, 0.015, 0.015)

    Limit may be a callable taking a number and returning
    the limited number.

    >>> at_least_one = lambda n: max(n, 1)
    >>> bd = BackoffDelay(delay=0.01, factor=2, limit=at_least_one)
    >>> next(bd)
    0.01
    >>> next(bd)
    1

    Pass a jitter to add or subtract seconds to the delay.

    >>> bd = BackoffDelay(jitter=0.01)
    >>> next(bd)
    0
    >>> next(bd)
    0.01

    Jitter may be a callable. To supply a non-deterministic jitter
    between -0.5 and 0.5, consider:

    >>> import random
    >>> jitter=functools.partial(random.uniform, -0.5, 0.5)
    >>> bd = BackoffDelay(jitter=jitter)
    >>> next(bd)
    0
    >>> 0 <= next(bd) <= 0.5
    True
    """

    factor: float = 1
    "Multiplier applied to delay"

    jitter: collections.abc.Callable[[], float]
    "Callable returning extra seconds to add to delay"

    @jaraco.functools.save_method_args
    def __init__(
        self,
        delay: float = 0,
        factor: float = 1,
        limit: collections.abc.Callable[[float], float] | float = float('inf'),
        jitter: collections.abc.Callable[[], float] | float = 0,
    ) -> None:
        self.delay = delay
        self.factor = factor
        if isinstance(limit, numbers.Number):
            limit_ = limit

            def limit_func(n: float, /) -> float:
                return max(0, min(limit_, n))

        else:
            # python/mypy#16946 or # python/mypy#13914
            limit_func: collections.abc.Callable[[float], float] = limit  # type: ignore[no-redef]
        self.limit = limit_func
        if isinstance(jitter, numbers.Number):
            jitter_ = jitter

            def jitter_func() -> float:
                return jitter_

        else:
            # python/mypy#16946 or # python/mypy#13914
            jitter_func: collections.abc.Callable[[], float] = jitter  # type: ignore[no-redef]

        self.jitter = jitter_func

    def __call__(self) -> None:
        time.sleep(next(self))

    def __next__(self) -> int | float:
        delay = self.delay
        self.bump()
        return delay

    def __iter__(self) -> Self:
        return self

    def bump(self) -> None:
        self.delay = self.limit(self.delay * self.factor + self.jitter())

    def reset(self) -> None:
        saved = self._saved___init__  # type: ignore[attr-defined]
        self.__init__(*saved.args, **saved.kwargs)  # type: ignore[misc]
